@@ -12,6 +12,7 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+
 """Ray-based executor for distributed machine learning workloads.
 
 This module provides the core execution framework for running distributed
@@ -43,7 +44,7 @@ Example:
     >>>
     >>> @ray.remote
     >>> def train_model(data):
-    ...     # Training logic here
+    ...
     ...     return trained_model
     >>>
     >>> tpu_config = TpuAcceleratorConfig(type="v4-8")
@@ -61,7 +62,7 @@ Example:
     >>> @execute_multislice_resumable(tpu_config)
     >>> @ray.remote
     >>> def distributed_train(slice_data):
-    ...     # Distributed training logic
+    ...
     ...     return slice_results
     >>>
     >>> results = distributed_train(training_data)
@@ -80,6 +81,7 @@ import functools
 import logging
 import os
 import time
+from collections.abc import Callable
 
 import ray
 from ray.exceptions import RayError
@@ -94,6 +96,7 @@ from .types import (
     JobPreempted,
     JobStatus,
     JobSucceeded,
+    SliceInfo,
     handle_ray_error,
 )
 
@@ -104,6 +107,17 @@ MEGASCALE_DEFAULT_PORT = 8081
 logger = logging.getLogger("ray")
 
 
+def resolve_maybe_refs(items):
+    """If items are Ray ObjectRefs, ray.get() them; otherwise return as-is."""
+    if not items:
+        return items
+    import ray
+
+    if all(isinstance(x, ray.ObjectRef) for x in items):
+        return ray.get(items)
+    return items
+
+
 class RayExecutor:
     """Core executor for Ray-based distributed workloads.
 
@@ -164,7 +178,7 @@ class RayExecutor:
             ...     return x * 2
             >>>
             >>> config = GpuAcceleratorConfig(count=1, type="v100")
-            >>> result = RayExecutor.execute(compute, config, x=10) # Returns JobStatus object
+            >>> result = RayExecutor.execute(compute, config, x=10)
         """
         assert getattr(accelerator_config, "pod_count", 1) == 1, (
             "Multi-slice workloads on TPUs should use 'execute_multislice'."
@@ -277,21 +291,21 @@ class RayExecutor:
         Example:
             >>> @ray.remote
             >>> def train_on_slice(data, slice_id):
-            ...     # Training logic for this slice
+            ...
             ...     return model_weights
             >>>
             >>> tpu_config = TpuAcceleratorConfig(type="v4-32", pod_count=4)
             >>>
-            >>> # Get flat list of results (default)
+            >>>
             >>> job_status = RayExecutor.execute_multislice(
             ...     train_on_slice,
             ...     tpu_config,
             ...     data=training_data
             ... )
             >>> if isinstance(job_status, JobSucceeded):
-            ...     flat_results = job_status.result  # Flat list from all hosts
+            ...     flat_results = job_status.result
+            >>>
             >>>
-            >>> # Get nested results by slice
             >>> job_status = RayExecutor.execute_multislice(
             ...     train_on_slice,
             ...     tpu_config,
@@ -299,7 +313,7 @@ class RayExecutor:
             ...     data=training_data
             ... )
             >>> if isinstance(job_status, JobSucceeded):
-            ...     results_by_slice = job_status.result  # List of lists
+            ...     results_by_slice = job_status.result
         """
         pool_manager = SlicePoolManager(tpu_type=accelerator_config.tpu_version)
         per_slice_futures = None
@@ -350,6 +364,8 @@ class RayExecutor:
                 host_futures = [
                     h.run_remote_fn.remote(
                         remote_fn,
+                        f_args=(),
+                        f_kwargs=kwargs,
                         runtime_env=accelerator_config.execution_env,
                         env=env_for_slice,
                     )
@@ -359,14 +375,11 @@ class RayExecutor:
 
             if flatten:
                 outer_refs = [f for sub in per_slice_futures for f in sub]
-                inner_refs = ray.get(outer_refs)
-                pending = list(inner_refs)
-                while pending:
-                    _, pending = ray.wait(pending, num_returns=1, timeout=10.0)
-                results = ray.get(inner_refs)
+                inner = ray.get(outer_refs)
+                results = resolve_maybe_refs(inner)
             else:
                 inner_by_slice = [ray.get(lst) for lst in per_slice_futures]
-                results = [ray.get(lst) for lst in inner_by_slice]
+                results = [resolve_maybe_refs(lst) for lst in inner_by_slice]
 
             return JobSucceeded(info, results)
 
@@ -392,8 +405,8 @@ class RayExecutor:
                         RayResources.cancel_all_futures(lst)
                 except Exception:
                     pass
-            info = JobInfo(accelerator_config.runtime_name, "running", accelerator_config.resource_name)
-            return JobFailed(info, e)
+            info2 = JobInfo(accelerator_config.runtime_name, "running", accelerator_config.resource_name)
+            return JobFailed(info2, e)
         finally:
             try:
                 pool_manager.drain_actor_pool()
@@ -448,7 +461,7 @@ class RayExecutor:
         Example:
             >>> @ray.remote
             >>> def long_running_task(data):
-            ...     # Task that might be preempted
+            ...
             ...     return process(data)
             >>>
             >>> config = TpuAcceleratorConfig(type="v4-8", preemptible=True)
@@ -573,12 +586,12 @@ class RayExecutor:
         Example:
             >>> @ray.remote
             >>> def distributed_training(data_shard):
-            ...     # Training logic for each slice
+            ...
             ...     return trained_weights
             >>>
             >>> tpu_config = TpuAcceleratorConfig(type="v4-32", pod_count=4)
             >>>
-            >>> # Get flat list of results
+            >>>
             >>> results = RayExecutor.execute_multislice_resumable(
             ...     distributed_training,
             ...     tpu_config,
@@ -587,7 +600,7 @@ class RayExecutor:
             ...     data_shard=sharded_data
             ... )
             >>>
-            >>> # Get nested results by slice
+            >>>
             >>> results_by_slice = RayExecutor.execute_multislice_resumable(
             ...     distributed_training,
             ...     tpu_config,
@@ -744,7 +757,7 @@ def execute_resumable(accelerator_config: AcceleratorConfigType):
         >>> def my_task(data):
         ...     return process(data)
         >>>
-        >>> result = my_task(input_data)  # Automatically retries on failure
+        >>> result = my_task(input_data)
     """
 
     def decorator(remote_fn: RemoteFunction):
@@ -791,7 +804,7 @@ def execute(accelerator_config: AcceleratorConfigType):
         >>> def gpu_task(tensor):
         ...     return tensor.cuda() * 2
         >>>
-        >>> result = gpu_task(my_tensor)  # Executes on GPU, no retry
+        >>> result = gpu_task(my_tensor)
     """
 
     def decorator(remote_fn: RemoteFunction):
@@ -837,7 +850,7 @@ def execute_multislice(accelerator_config: AcceleratorConfigType):
         >>> def parallel_compute(data_shard):
         ...     return compute_result(data_shard)
         >>>
-        >>> results = parallel_compute(sharded_data)  # Returns list from 4 slices
+        >>> results = parallel_compute(sharded_data)
     """
 
     def decorator(remote_fn: RemoteFunction):
@@ -881,10 +894,10 @@ def execute_multislice_resumable(accelerator_config: AcceleratorConfigType):
         >>> @execute_multislice_resumable(tpu_config)
         >>> @ray.remote
         >>> def resilient_training(data_batch):
-        ...     # Long-running training that might be preempted
+        ...
         ...     return train_model(data_batch)
         >>>
-        >>> results = resilient_training(training_data)  # Auto-retries on failure
+        >>> results = resilient_training(training_data)
     """
 
     def decorator(remote_fn: RemoteFunction):
@@ -899,3 +912,529 @@ def execute_multislice_resumable(accelerator_config: AcceleratorConfigType):
         return wrapper
 
     return decorator
+
+
+def _build_multislice_envs(
+    slice_infos: list[SliceInfo],
+    base_env: dict[str, str],
+    coord_port: int,
+) -> list[dict[str, str]]:
+    """Build environment variables for multi-slice TPU execution.
+
+    Creates environment variable dictionaries for each TPU slice in a multi-slice
+    setup, including MegaScale coordination variables required for inter-slice
+    communication.
+
+    Args:
+        slice_infos: List of SliceInfo objects containing metadata about each
+            TPU slice, including IP addresses and slice names.
+        base_env: Base environment variables to include for all slices.
+        coord_port: Port number for the MegaScale coordinator service.
+
+    Returns:
+        List of environment variable dictionaries, one per slice. Each dictionary
+        contains:
+        - All base environment variables
+        - MegaScale coordination variables (if multi-slice):
+            - MEGASCALE_COORDINATOR_ADDRESS: IP:port of coordinator
+            - MEGASCALE_NUM_SLICES: Total number of slices
+            - MEGASCALE_PORT: Coordinator port
+            - MEGASCALE_SLICE_ID: This slice's ID (0-indexed)
+            - EXECUTOR_CALL_INDEX: Worker index within slice
+            - EXECUTOR_CALL_SLICE: Same as MEGASCALE_SLICE_ID
+            - TPU_SLICE_NAME: Name of this TPU slice
+
+    Note:
+        The first slice's IP address is used as the coordinator address.
+        All values are converted to strings for environment compatibility.
+    """
+    envs = []
+    coord_ip = slice_infos[0].ip_address
+    for i, si in enumerate(slice_infos):
+        env = dict(base_env)
+        if len(slice_infos) > 1:
+            env.update(
+                MEGASCALE_COORDINATOR_ADDRESS=f"{coord_ip}:{coord_port}",
+                MEGASCALE_NUM_SLICES=str(len(slice_infos)),
+                MEGASCALE_PORT=str(coord_port),
+                MEGASCALE_SLICE_ID=str(i),
+                EXECUTOR_CALL_INDEX="0",
+                EXECUTOR_CALL_SLICE=str(i),
+                TPU_SLICE_NAME=si.slice_name,
+            )
+        envs.append({str(k): str(v) for k, v in env.items()})
+    return envs
+
+
+class TpuRemoteManager:
+    """Session-based manager for TPU remote execution across multiple slices.
+
+    Provides a high-level interface for managing TPU resources and executing
+    functions or class methods across multiple TPU slices. Each execution runs
+    in a short-lived worker process to avoid TPU handle leaks.
+
+    Key features:
+    - Automatic slice scaling and warming
+    - Persistent DeviceHostActors for efficient execution
+    - Broadcasting of function/method calls across all hosts
+    - Ephemeral worker processes for each execution
+    - MegaScale coordination for multi-slice workloads
+
+    Attributes:
+        tpu_version: TPU version string (e.g., 'v4', 'v5p')
+        pod_count: Number of TPU pods/slices to use
+        base_env: Base environment variables for all workers
+        runtime_env: Ray runtime environment configuration
+        coord_port: Port for MegaScale coordinator (default 8081)
+
+    Example:
+        >>> manager = TpuRemoteManager(
+        ...     tpu_version='v4',
+        ...     pod_count=2,
+        ...     base_env={'MY_VAR': 'value'}
+        ... )
+        >>> manager.ensure()
+        >>> results = manager.run_function(my_func, arg1, arg2)
+        >>> manager.close()
+    """
+
+    def __init__(
+        self,
+        tpu_version: str,
+        pod_count: int,
+        *,
+        base_env: dict[str, str] | None = None,
+        runtime_env: dict | None = None,
+        coord_port: int = 8081,
+    ):
+        self.tpu_version = tpu_version
+        self.pod_count = int(pod_count)
+        self.base_env = dict(base_env or {})
+        self.runtime_env = dict(runtime_env or {})
+        self.coord_port = int(coord_port)
+
+        self._pool = SlicePoolManager(tpu_type=tpu_version)
+        self._members = None
+        self._slice_infos: list[SliceInfo] = []
+        self._per_slice_envs: list[dict[str, str]] = []
+        self._host_handles_by_slice: list[list[ray.actor.ActorHandle]] = []
+
+    def ensure(self):
+        """Initialize and prepare TPU resources if not already done.
+
+        This method:
+        1. Scales the slice pool to the requested pod_count
+        2. Prepares all slices (creates placement groups, etc.)
+        3. Ensures DeviceHostActors are created on each host
+        4. Builds per-slice environment variables including MegaScale config
+        5. Caches host actor handles for efficient execution
+
+        Raises:
+            RuntimeError: If no SliceActors are available after scaling.
+
+        Note:
+            This method is idempotent - calling it multiple times has no
+            additional effect after the first successful call.
+        """
+        if self._members is not None:
+            return
+
+        self._pool.scale_multislice(self.pod_count)
+        self._pool.prepare_all_slices()
+
+        members = self._pool.get_all_pool_members()
+        if not members:
+            raise RuntimeError("No SliceActors available after scaling.")
+        ray.get([m.actor.ensure_host_pool.remote() for m in members])
+
+        slice_infos = ray.get([m.actor.get_info.remote() for m in members])
+
+        base_env = dict(self.base_env)
+        base_env.update(
+            TPU_VERSION=self.tpu_version,
+            TPU_POD_COUNT=str(len(members)),
+        )
+        self._per_slice_envs = _build_multislice_envs(slice_infos, base_env, self.coord_port)
+
+        host_handles_by_slice = ray.get([m.actor.get_all_actors_in_pool.remote() for m in members])
+
+        self._members = members
+        self._slice_infos = slice_infos
+        self._host_handles_by_slice = host_handles_by_slice
+
+    def close(self):
+        """Clean up all TPU resources and reset state.
+
+        Performs cleanup in the following order:
+        1. Cancels any current work on all host actors
+        2. Drains the actor pool (terminates all actors)
+        3. Resets internal state
+
+        Note:
+            This method is safe to call multiple times and will suppress
+            any errors during cleanup to ensure resources are freed.
+        """
+
+        try:
+            if self._members:
+                for m in self._members:
+                    try:
+                        hosts = ray.get(m.actor.get_all_actors_in_pool.remote())
+                        ray.get([h.cancel_current.remote() for h in hosts], timeout=10)
+                    except Exception:
+                        pass
+        except Exception:
+            pass
+        try:
+            self._pool.drain_actor_pool()
+        finally:
+            self._members = None
+            self._slice_infos = []
+            self._per_slice_envs = []
+            self._host_handles_by_slice = []
+
+    def run_function(self, fn: Callable, *args, flatten: bool = True, **kwargs):
+        """Execute a Python function on every TPU host across all slices.
+
+        Runs the provided function in ephemeral worker processes on each host
+        to avoid TPU handle leaks. The function is executed with the same
+        arguments on every host.
+
+        Args:
+            fn: Python callable to execute on each host. Should be pickleable.
+            *args: Positional arguments to pass to the function.
+            flatten: If True, returns a flat list of results from all hosts.
+                If False, returns nested lists where outer list represents
+                slices and inner lists contain results from hosts within
+                each slice.
+            **kwargs: Keyword arguments to pass to the function.
+
+        Returns:
+            If flatten=True: List of results from all hosts (flattened).
+            If flatten=False: List of lists, where each inner list contains
+                results from hosts within a slice.
+
+        Note:
+            Each execution runs in a new process to avoid TPU handle leaks.
+            The function must be pickleable for Ray serialization.
+        """
+        self.ensure()
+
+        outer_refs_per_slice = []
+        for sl, hosts in enumerate(self._host_handles_by_slice):
+            env = self._per_slice_envs[sl]
+
+            outer = [
+                h.run_remote_fn.remote(
+                    fn,
+                    f_args=args,
+                    f_kwargs=kwargs,
+                    runtime_env=self.runtime_env,
+                    env=env,
+                )
+                for h in hosts
+            ]
+            outer_refs_per_slice.append(outer)
+
+        inner_per_slice = [ray.get(outer) for outer in outer_refs_per_slice]
+        if flatten:
+            flat_refs = [ref for refs in inner_per_slice for ref in refs]
+            return ray.get(flat_refs)
+        else:
+            return [ray.get(refs) for refs in inner_per_slice]
+
+    def run_class_method(
+        self,
+        cls_obj: type,
+        init_args: tuple,
+        init_kwargs: dict,
+        method_name: str,
+        *call_args,
+        flatten: bool = True,
+        **call_kwargs,
+    ):
+        """Execute a class method on every TPU host across all slices.
+
+        Instantiates the class and calls the specified method in ephemeral
+        worker processes on each host. The class is reconstructed inside
+        each worker to ensure clean TPU state.
+
+        Args:
+            cls_obj: The class object (not an instance). Must be pickleable.
+            init_args: Positional arguments for class initialization.
+            init_kwargs: Keyword arguments for class initialization.
+            method_name: Name of the method to call on the instantiated object.
+            *call_args: Positional arguments for the method call.
+            flatten: If True, returns a flat list of results from all hosts.
+                If False, returns nested lists by slice.
+            **call_kwargs: Keyword arguments for the method call.
+
+        Returns:
+            If flatten=True: List of results from all hosts (flattened).
+            If flatten=False: List of lists, where each inner list contains
+                results from hosts within a slice.
+
+        Example:
+            >>> results = manager.run_class_method(
+            ...     MyModel,
+            ...     init_args=(config,),
+            ...     init_kwargs={'checkpoint': 'path/to/ckpt'},
+            ...     method_name='train',
+            ...     epochs=10,
+            ...     batch_size=32
+            ... )
+
+        Note:
+            The class is instantiated fresh in each worker process,
+            avoiding TPU handle leaks from persistent objects.
+        """
+        self.ensure()
+
+        def _class_method_entry(cls_obj, i_args, i_kwargs, mname, c_args, c_kwargs):
+            obj = cls_obj(*i_args, **i_kwargs)
+            return getattr(obj, mname)(*c_args, **c_kwargs)
+
+        outer_refs_per_slice = []
+        for sl, hosts in enumerate(self._host_handles_by_slice):
+            env = self._per_slice_envs[sl]
+            payload = (cls_obj, init_args, init_kwargs, method_name, call_args, call_kwargs)
+            outer = [
+                h.run_remote_fn.remote(
+                    _class_method_entry,
+                    f_args=payload,
+                    runtime_env=self.runtime_env,
+                    env=env,
+                )
+                for h in hosts
+            ]
+            outer_refs_per_slice.append(outer)
+
+        inner_per_slice = [ray.get(outer) for outer in outer_refs_per_slice]
+        if flatten:
+            flat_refs = [ref for refs in inner_per_slice for ref in refs]
+            return ray.get(flat_refs)
+        else:
+            return [ray.get(refs) for refs in inner_per_slice]
+
+
+class _BoundRemoteClass:
+    """Bound handle for remote class execution on TPU hosts.
+
+    Internal class that provides a proxy interface for calling methods on
+    a class that will be instantiated and executed on TPU hosts. Supports
+    both synchronous execution (returning values) and asynchronous execution
+    (returning Ray ObjectRefs).
+
+    This class is created by the tpu_remote decorator when applied to classes.
+
+    Attributes:
+        _manager: TpuRemoteManager instance for resource management
+        _cls: The class object to instantiate on remote hosts
+        _init_args: Positional arguments for class initialization
+        _init_kwargs: Keyword arguments for class initialization
+        _flatten: Whether to flatten results from multiple hosts
+    """
+
+    def __init__(self, manager: TpuRemoteManager, cls: type, init_args: tuple, init_kwargs: dict, flatten: bool):
+        self._manager = manager
+        self._cls = cls
+        self._init_args = init_args
+        self._init_kwargs = init_kwargs
+        self._flatten = flatten
+
+    def __getattr__(self, name: str):
+        def _call_method(*args, **kwargs):
+            return self._manager.run_class_method(
+                self._cls,
+                self._init_args,
+                self._init_kwargs,
+                name,
+                *args,
+                flatten=self._flatten,
+                **kwargs,
+            )
+
+        _call_method.remote = self._remote(name)
+        return _call_method
+
+    def _remote(self, name: str):
+        def _remote_impl(*args, **kwargs):
+            def _class_method_entry(cls_obj, i_args, i_kwargs, mname, c_args, c_kwargs):
+                obj = cls_obj(*i_args, **i_kwargs)
+                return getattr(obj, mname)(*c_args, **c_kwargs)
+
+            self._manager.ensure()
+            outer_refs_per_slice = []
+            for sl, hosts in enumerate(self._manager._host_handles_by_slice):
+                env = self._manager._per_slice_envs[sl]
+                payload = (self._cls, self._init_args, self._init_kwargs, name, args, kwargs)
+                outer = [
+                    h.run_remote_fn.remote(
+                        _class_method_entry,
+                        f_args=payload,
+                        runtime_env=self._manager.runtime_env,
+                        env=env,
+                    )
+                    for h in hosts
+                ]
+                outer_refs_per_slice.append(outer)
+
+            inner_per_slice = [ray.get(outer) for outer in outer_refs_per_slice]
+            if self._flatten:
+                return [ref for refs in inner_per_slice for ref in refs]
+            else:
+                return inner_per_slice
+
+        return _remote_impl
+
+    def close(self):
+        """Close the manager and clean up TPU resources.
+
+        Delegates to the underlying TpuRemoteManager to terminate all actors
+        and free TPU resources. Safe to call multiple times.
+        """
+        self._manager.close()
+
+
+class _RemoteClassWrapper:
+    """Wrapper that converts a regular class into a TPU-remote class.
+
+    Internal class used by the tpu_remote decorator to wrap classes for
+    remote execution. When the wrapped class is instantiated, it returns
+    a _BoundRemoteClass that broadcasts method calls to TPU hosts.
+
+    Attributes:
+        _manager: TpuRemoteManager instance for resource management
+        _flatten: Whether to flatten results from multiple hosts
+    """
+
+    def __init__(self, manager: TpuRemoteManager, flatten: bool):
+        self._manager = manager
+        self._flatten = flatten
+
+    def __call__(self, cls: type):
+        @functools.wraps(cls)
+        def ctor(*args, **kwargs):
+            return _BoundRemoteClass(self._manager, cls, args, kwargs, self._flatten)
+
+        return ctor
+
+
+class _RemoteFunctionWrapper:
+    """Wrapper that converts a regular function into a TPU-remote function.
+
+    Internal class used by the tpu_remote decorator to wrap functions for
+    remote execution. The wrapped function broadcasts calls to all TPU hosts
+    and supports both synchronous and asynchronous execution modes.
+
+    Attributes:
+        _manager: TpuRemoteManager instance for resource management
+        _flatten: Whether to flatten results from multiple hosts
+    """
+
+    def __init__(self, manager: TpuRemoteManager, flatten: bool):
+        self._manager = manager
+        self._flatten = flatten
+
+    def __call__(self, fn: Callable):
+        @functools.wraps(fn)
+        def wrapper(*args, **kwargs):
+            return self._manager.run_function(fn, *args, flatten=self._flatten, **kwargs)
+
+        def _remote(*args, **kwargs):
+            self._manager.ensure()
+            outer_refs_per_slice = []
+            for sl, hosts in enumerate(self._manager._host_handles_by_slice):
+                env = self._manager._per_slice_envs[sl]
+                outer = [
+                    h.run_remote_fn.remote(
+                        fn,
+                        f_args=args,
+                        f_kwargs=kwargs,
+                        runtime_env=self._manager.runtime_env,
+                        env=env,
+                    )
+                    for h in hosts
+                ]
+                outer_refs_per_slice.append(outer)
+            inner_per_slice = [ray.get(outer) for outer in outer_refs_per_slice]
+            if self._flatten:
+                return [ref for refs in inner_per_slice for ref in refs]
+            else:
+                return inner_per_slice
+
+        wrapper.remote = _remote  # type: ignore[attr-defined]
+        return wrapper
+
+
+def device_remote(*, accelerator_config: TpuAcceleratorConfig, flatten: bool = True):
+    """Decorator for TPU-remote execution of functions or classes.
+
+    Transforms a regular Python function or class into a TPU-remote version
+    that automatically broadcasts execution across all TPU hosts in the
+    specified configuration. Each execution runs in an ephemeral worker
+    process to avoid TPU handle leaks.
+
+    Args:
+        accelerator_config: TPU configuration specifying version, pod count,
+            and other execution parameters.
+        flatten: If True (default), returns flat list of results from all
+            hosts. If False, returns nested lists where outer list represents
+            slices and inner lists contain results from hosts within each slice.
+
+    Returns:
+        Decorator function that wraps the target function or class.
+
+    Example for functions:
+        >>> @device_remote(accelerator_config=tpu_config)
+        >>> def compute(x, y):
+        ...     return jax.numpy.dot(x, y)
+        >>>
+        >>>
+        >>> results = compute(array1, array2)
+        >>>
+        >>>
+        >>> refs = compute.remote(array1, array2)
+        >>> results = ray.get(refs)
+
+    Example for classes:
+        >>> @tpu_remote(accelerator_config=tpu_config)
+        >>> class Model:
+        ...     def __init__(self, config):
+        ...         self.config = config
+        ...
+        ...     def train(self, data):
+        ...
+        ...         return metrics
+        >>>
+        >>> model = Model(config)
+        >>> metrics = model.train(data)
+        >>> refs = model.train.remote(data)
+
+    Note:
+        - Functions/classes must be pickleable for Ray serialization
+        - Each method call creates new worker processes on TPU hosts
+        - The decorator manages slice scaling and resource allocation
+        - Call model.close() to clean up resources when done
+    """
+    full_runtime_env = dict(accelerator_config.execution_env or {})
+    env_vars = dict(full_runtime_env.get("env_vars", {}))
+    full_runtime_env["env_vars"] = env_vars
+
+    mgr = TpuRemoteManager(
+        tpu_version=accelerator_config.tpu_version,
+        pod_count=accelerator_config.pod_count,
+        base_env=env_vars,
+        runtime_env=full_runtime_env,
+        coord_port=8081,
+    )
+
+    def decorator(obj: Callable | type):
+        if isinstance(obj, type):
+            return _RemoteClassWrapper(mgr, flatten)(obj)
+        elif callable(obj):
+            return _RemoteFunctionWrapper(mgr, flatten)(obj)
+        else:
+            raise TypeError("tpu_remote can only decorate a function or a class.")
+
+    return decorator
