@@ -12,28 +12,45 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+
 """Resource pool management for distributed Ray actors.
 
-This module provides abstractions for managing pools of Ray actors,
-particularly focused on TPU/GPU slice management for distributed computing.
-It includes health monitoring, automatic scaling, and resource lifecycle
-management.
+This module provides comprehensive abstractions for managing pools of Ray actors,
+with specialized focus on TPU/GPU slice management for distributed computing.
+It includes health monitoring, automatic scaling, resource lifecycle management,
+and placement group coordination for optimal resource allocation.
 
 Key Components:
     - **ActorPoolMember**: Wrapper for actor handles with metadata
     - **ResourcePoolManager**: Abstract base for managing actor pools
-    - **SlicePoolManager**: Specialized manager for TPU/GPU slices
+    - **SlicePoolManager**: Specialized manager for TPU/GPU slices with placement groups
     - **SliceActor**: Ray actor for managing individual compute slices
+    - **DeviceHostActor**: Ray actor for managing individual TPU hosts within slices
+
+Resource Management Features:
+    - Placement group coordination with STRICT_SPREAD strategy
+    - Automatic resource request handling through Ray autoscaler
+    - Health monitoring with graceful shutdown sequences
+    - Robust error handling with actor restart capabilities
+    - Slot-based actor allocation for deterministic placement
+
+Environment Variables:
+    - **EFORMER_SCALE_POLL_S**: Scaling operation polling interval (default: "30")
+    - **EFORMER_SCALE_ADD_TIMEOUT_S**: Timeout for adding new actors (default: "604800")
 
 Example:
-    Managing a multi-slice TPU configuration:
+    Managing a multi-slice TPU configuration with placement groups:
 
     >>> from eformer.executor.ray import SlicePoolManager
     >>>
+    >>> # Create manager with automatic placement group coordination
     >>> manager = SlicePoolManager(tpu_type="v4-8")
     >>> manager.scale_multislice(num_slices=4)
     >>> actors = manager.get_all_actors_in_pool()
-    >>> manager.drain_actor_pool()
+    >>>
+    >>> # Placement groups are automatically managed
+    >>> manager.prepare_all_slices()
+    >>> manager.drain_actor_pool()  # Cleans up placement groups
 """
 
 from __future__ import annotations
@@ -50,7 +67,6 @@ import ray
 import requests
 from ray.actor import ActorHandle
 from ray.autoscaler.sdk import request_resources
-from ray.remote_function import RemoteFunction as _RF
 from ray.util.placement_group import placement_group, remove_placement_group
 from ray.util.scheduling_strategies import NodeAffinitySchedulingStrategy, PlacementGroupSchedulingStrategy
 
@@ -79,10 +95,10 @@ class InsufficientSlicesError(RuntimeError):
     Example:
         >>> manager = SlicePoolManager(tpu_type="v4-32")
         >>> try:
-        ...     manager.scale_multislice([4, 8])  # Request 8 slices, fallback to 4
+        ...     manager.scale_multislice([4, 8])
         ... except InsufficientSlicesError as e:
         ...     print(f"Could not allocate TPU slices: {e}")
-        ...     # Handle fallback or retry logic
+        ...
     """
 
     pass
@@ -176,10 +192,10 @@ class ResourcePoolManager(Generic[ActorInfoT]):
         ref_map = {m: m.actor.healthy.remote() for m in self._actor_pool}
         refs = list(ref_map.values())
 
-        done, pending = ray.wait(refs, num_returns=len(refs), timeout=HEALTH_CHECK_TIMEOUT_S)
+        done, _ = ray.wait(refs, num_returns=len(refs), timeout=HEALTH_CHECK_TIMEOUT_S)
 
         done_set = set(done)
-        healthy: list[ActorPoolMember[ActorInfoT]] = []
+        healthy: list[ActorPoolMember[HostInfo]] = []
 
         for member, ref in ref_map.items():
             name = self.get_actor_name_from_actor_info(member.actor_info)
@@ -190,19 +206,19 @@ class ResourcePoolManager(Generic[ActorInfoT]):
                     else:
                         logger.warning(f"Actor {name} reported unhealthy; killing")
                         try:
-                            ray.kill(member.actor)
+                            ray.kill(member.actor, no_restart=True)
                         except Exception:
                             pass
                 except Exception as e:
                     logger.warning(f"Actor {name} health check exception ({e}); killing")
                     try:
-                        ray.kill(member.actor)
+                        ray.kill(member.actor, no_restart=True)
                     except Exception:
                         pass
             else:
                 logger.warning(f"Actor {name} health timeout; killing")
                 try:
-                    ray.kill(member.actor)
+                    ray.kill(member.actor, no_restart=True)
                 except Exception:
                     pass
 
@@ -239,7 +255,7 @@ class ResourcePoolManager(Generic[ActorInfoT]):
             except Exception as e:
                 logger.warning(f"SliceActor failed to start in time: {e}; killing actor")
                 try:
-                    ray.kill(actor)
+                    ray.kill(actor, no_restart=True)
                 except Exception:
                     pass
 
@@ -259,7 +275,7 @@ class ResourcePoolManager(Generic[ActorInfoT]):
                     ray.get(member.actor.shutdown.remote(), timeout=5)
                 except Exception:
                     pass
-                ray.kill(member.actor)
+                ray.kill(member.actor, no_restart=True)
                 logger.info(f"Removed actor {name}")
             except Exception as e:
                 logger.error(f"Failed to kill actor {name}: {e}")
@@ -304,7 +320,7 @@ class ResourcePoolManager(Generic[ActorInfoT]):
         for member in self._actor_pool:
             name = self.get_actor_name_from_actor_info(member.actor_info)
             try:
-                ray.kill(member.actor)
+                ray.kill(member.actor, no_restart=True)
                 logger.info(f"Killed actor {name}")
             except Exception as e:
                 logger.error(f"Failed to kill actor {name}: {e}")
@@ -462,6 +478,8 @@ class DeviceHostActor:
         self,
         remote_fn,
         *,
+        f_args: tuple = (),
+        f_kwargs: dict | None = None,
         runtime_env: dict | None = None,
         env: dict | None = None,
         num_cpus: float = 8.0,
@@ -501,32 +519,53 @@ class DeviceHostActor:
         if not self.healthy():
             raise RuntimeError(f"Host {self.host_id} unhealthy or preempted")
 
-        if not isinstance(remote_fn, _RF):
-            remote_fn = ray.remote(max_calls=1)(remote_fn)
-        elif remote_fn._default_options.get("max_calls") is None:
-            raise ValueError("remote_fn must have max_calls=1 for TPU workloads")
-
         if self._awaitable:
             self._cancel_tasks_and_wait([self._awaitable])
 
         self._hacky_remove_tpu_lockfile()
+
         host_env = {"TPU_HOST_ID": str(self.host_id), "TPU_SLICE_NAME": self.slice_name}
         if self.num_devices:
             host_env["TPU_NUM_DEVICES"] = str(self.num_devices)
         merged_runtime_env = self._merge_runtime_env(runtime_env, {**host_env, **(env or {})})
+
         resources = dict(extra_resources or {})
 
-        if self.num_devices:
-            resources["TPU"] = self.num_devices
+        py_fn = None
+        try:
+            from ray.remote_function import RemoteFunction as _RF
 
-        self._awaitable = remote_fn.options(
+            if isinstance(remote_fn, _RF):
+                py_fn = remote_fn._function
+            else:
+                py_fn = remote_fn
+        except Exception:
+            py_fn = remote_fn
+
+        f_kwargs = f_kwargs or {}
+
+        @ray.remote(max_calls=1)
+        def _runner(fn, args, kwargs):
+            try:
+                return fn(*args, **kwargs)
+            finally:
+                try:
+                    import jax.distributed as jdist
+
+                    jdist.shutdown()
+                except Exception:
+                    pass
+
+        self._awaitable = _runner.options(
             scheduling_strategy=NodeAffinitySchedulingStrategy(self._node_id, soft=False),
             resources=resources or None,
             num_cpus=num_cpus,
             num_gpus=0,
-            memory=memory_bytes,
+            memory=int(memory_bytes),
             runtime_env=merged_runtime_env,
-        ).remote()
+            max_retries=0,
+        ).remote(py_fn, f_args, f_kwargs)
+
         return self._awaitable
 
     def shutdown(self) -> None:
@@ -642,13 +681,8 @@ class SliceActor:
         return DeviceHostActor.options(
             num_cpus=0,
             scheduling_strategy=NodeAffinitySchedulingStrategy(node_id=node_id, soft=False),
-            lifetime="detached",
             name=f"{self._slice_info.slice_name}-host-{host_id}-{uuid4().hex[:8]}",
-        ).remote(
-            host_id,
-            self._slice_info.slice_name,
-            num_devices_for_host,
-        )
+        ).remote(host_id, self._slice_info.slice_name, num_devices_for_host)
 
     def _initialize_slice_info(self) -> None:
         """Initialize slice information from TPU environment.
@@ -661,6 +695,12 @@ class SliceActor:
 
             slice_name = ray_tpu.get_current_pod_name()
             num_hosts = int(ray_tpu.get_current_pod_worker_count())
+            if os.getenv("EFORMER_MODERATE", "1") == "1":
+                available_hosts = ray.cluster_resources().get(slice_name, None)
+                if available_hosts is not None and num_hosts > available_hosts:
+                    available_hosts = int(available_hosts)
+                    logger.info(f"setting {num_hosts=} to {available_hosts}")
+                    num_hosts = available_hosts
             ip_address = ray.util.get_node_ip_address()
 
             num_accelerators_per_host = None
@@ -791,7 +831,7 @@ class SliceActor:
                 scheduling_strategy=PlacementGroupSchedulingStrategy(
                     self._host_placement_group,
                     placement_group_bundle_index=i,
-                    placement_group_capture_child_tasks=True,
+                    placement_group_capture_child_tasks=False,
                 )
             ).remote()
             for i in range(self._slice_info.num_hosts)
@@ -871,7 +911,7 @@ class SliceActor:
         ref_map = {m: m.actor.healthy.remote() for m in self._actor_pool}
         refs = list(ref_map.values())
 
-        done, pending = ray.wait(refs, num_returns=len(refs), timeout=HEALTH_CHECK_TIMEOUT_S)
+        done, _ = ray.wait(refs, num_returns=len(refs), timeout=HEALTH_CHECK_TIMEOUT_S)
 
         done_set = set(done)
         healthy: list[ActorPoolMember[ActorInfoT]] = []
@@ -885,19 +925,19 @@ class SliceActor:
                     else:
                         logger.warning(f"Actor {name} reported unhealthy; killing")
                         try:
-                            ray.kill(member.actor)
+                            ray.kill(member.actor, no_restart=True)
                         except Exception:
                             pass
                 except Exception as e:
                     logger.warning(f"Actor {name} health check exception ({e}); killing")
                     try:
-                        ray.kill(member.actor)
+                        ray.kill(member.actor, no_restart=True)
                     except Exception:
                         pass
             else:
                 logger.warning(f"Actor {name} health timeout; killing")
                 try:
-                    ray.kill(member.actor)
+                    ray.kill(member.actor, no_restart=True)
                 except Exception:
                     pass
 
@@ -946,7 +986,7 @@ class SliceActor:
                 except Exception as e:
                     logger.error(f"Failed to start host actor: {e}")
                     try:
-                        ray.kill(actor)
+                        ray.kill(actor, no_restart=True)
                     except Exception:
                         pass
 
@@ -964,7 +1004,7 @@ class SliceActor:
                     ray.get(member.actor.shutdown.remote(), timeout=5)
                 except Exception:
                     pass
-                ray.kill(member.actor)
+                ray.kill(member.actor, no_restart=True)
                 logger.info(f"Removed actor {name}")
             except Exception as e:
                 logger.error(f"Failed to kill actor {name}: {e}")
@@ -1009,7 +1049,7 @@ class SliceActor:
         for member in self._actor_pool:
             name = self.get_actor_name_from_actor_info(member.actor_info)
             try:
-                ray.kill(member.actor)
+                ray.kill(member.actor, no_restart=True)
                 logger.info(f"Killed actor {name}")
             except Exception as e:
                 logger.error(f"Failed to kill actor {name}: {e}")
@@ -1017,6 +1057,18 @@ class SliceActor:
         self._actor_pool = []
 
     def _await_all_hosts_healthy(self, timeout_s: int = 60, poll_s: float = 2.0) -> bool:
+        """Wait for all hosts in the pool to become healthy.
+
+        Polls the health status of all host actors until they all report
+        healthy or the timeout is reached.
+
+        Args:
+            timeout_s: Maximum time to wait for hosts to become healthy (default: 60).
+            poll_s: Interval between health checks in seconds (default: 2.0).
+
+        Returns:
+            True if all hosts became healthy within the timeout, False otherwise.
+        """
         deadline = time.time() + timeout_s
         while time.time() < deadline:
             statuses = ray.get([m.actor.healthy.remote() for m in self._actor_pool])
@@ -1025,7 +1077,14 @@ class SliceActor:
             time.sleep(poll_s)
         return False
 
-    def run_remote_fn(self, remote_fn, runtime_env: dict | None = None, env: dict | None = None):
+    def run_remote_fn(
+        self,
+        remote_fn,
+        runtime_env: dict | None = None,
+        env: dict | None = None,
+        f_args: tuple = (),
+        f_kwargs: dict | None = None,
+    ):
         """Execute a remote function on all hosts in this slice.
 
         Ensures all hosts are ready, then launches the function on each host
@@ -1061,8 +1120,16 @@ class SliceActor:
             self._await_all_hosts_healthy(timeout_s=int(os.getenv("EFORMER_HOST_HEALTH_WAIT_S", "60")))
         except Exception:
             pass
+
         futures = [
-            member.actor.run_remote_fn.remote(remote_fn, runtime_env=runtime_env, env=env) for member in self._actor_pool
+            member.actor.run_remote_fn.remote(
+                remote_fn,
+                f_args=f_args,
+                f_kwargs=f_kwargs,
+                runtime_env=runtime_env,
+                env=env,
+            )
+            for member in self._actor_pool
         ]
         return futures
 
@@ -1119,6 +1186,8 @@ class SlicePoolManager(ResourcePoolManager[SliceInfo]):
         self._tpu_type = tpu_type
         self._last_scale_ts: float | None = None
         self._last_scale_check_ts: float | None = None
+        self._head_pg = None
+        self._head_pg_target = 0
 
     def get_actor_pool_name(self) -> str:
         """Get a human-readable name for this actor pool.
@@ -1169,8 +1238,8 @@ class SlicePoolManager(ResourcePoolManager[SliceInfo]):
             InsufficientSlicesError: If none of the requested sizes can be achieved.
 
         Example:
-            >>> manager.scale_multislice(4)  # Exactly 4 slices
-            >>> manager.scale_multislice([2, 4, 8])  # Try 8, fall back to 4 or 2
+            >>> manager.scale_multislice(4)
+            >>> manager.scale_multislice([2, 4, 8])
 
         Note:
             - Requests TPU head resources from Ray autoscaler.
@@ -1218,15 +1287,44 @@ class SlicePoolManager(ResourcePoolManager[SliceInfo]):
             Called automatically by execute_multislice before running tasks.
             Essential for proper multi-host coordination within each slice.
         """
-        slice_infos: list[SliceInfo] = ray.get([m.actor.get_info.remote() for m in self._actor_pool])
-        all_bundles = []
 
-        for info in slice_infos:
-            all_bundles.extend([{"CPU": 0, info.slice_name: 1}] * info.num_hosts)
-        if all_bundles:
-            request_resources(bundles=all_bundles)
+        if os.getenv("EFORMER_SAFE_GATHER", "1") == "1":
+            slice_infos = []
+            good_members = []
+            for m in self._actor_pool:
+                try:
+                    si = ray.get(m.actor.get_info.remote(), timeout=30)
+                    slice_infos.append(si)
+                    good_members.append(m)
+                except Exception as e:
+                    try:
+                        ray.kill(m.actor, no_restart=True)
+                    except Exception:
+                        pass
+                    logger.warning(f"Pruned dead SliceActor during prepare_all_slices: {e}")
 
-        ray.get([m.actor.prepare_hosts.remote() for m in self._actor_pool])
+            if len(good_members) != len(self._actor_pool):
+                self._actor_pool = good_members
+                if not self._actor_pool:
+                    raise RuntimeError("No SliceActors available after pruning.")
+
+            all_bundles = []
+            for info in slice_infos:
+                all_bundles.extend([{"CPU": 0, info.slice_name: 1}] * info.num_hosts)
+            if all_bundles:
+                request_resources(bundles=all_bundles)
+
+            ray.get([m.actor.prepare_hosts.remote() for m in self._actor_pool])
+        else:
+            slice_infos: list[SliceInfo] = ray.get([m.actor.get_info.remote() for m in self._actor_pool])
+            all_bundles = []
+
+            for info in slice_infos:
+                all_bundles.extend([{"CPU": 0, info.slice_name: 1}] * info.num_hosts)
+            if all_bundles:
+                request_resources(bundles=all_bundles)
+
+            ray.get([m.actor.prepare_hosts.remote() for m in self._actor_pool])
 
     def should_scale_up_multislice(self, valid_sizes: Sequence[int]) -> bool:
         """Check if pool should scale up to a larger size.
@@ -1365,57 +1463,163 @@ class SlicePoolManager(ResourcePoolManager[SliceInfo]):
             - Progress is logged throughout the scaling process
         """
         current = len(self._actor_pool)
+
         if current >= desired_num_actors:
             return
-        num_to_add = desired_num_actors - current
+
+        if current != 0:
+            logger.info("Recreating head PG due to non-zero current pool (align bundles with slots)")
+            self.drain_actor_pool()
+            current = 0
+
         logger.info(f"Scaling up pool {self.get_actor_pool_name()} from {current} to {desired_num_actors}")
 
-        info_ref_to_actor: dict[ray.ObjectRef, ActorHandle] = {}
-        for _ in range(num_to_add):
-            actor = self.create_actor()
-            info_ref = actor.get_info.remote()
-            info_ref_to_actor[info_ref] = actor
+        self._ensure_head_pg(desired_num_actors)
+
+        slot_to_actor: dict[int, ActorHandle] = {}
+        slot_to_info_ref: dict[int, ray.ObjectRef] = {}
+
+        def _start_for_slot(slot: int):
+            actor = SliceActor.options(
+                num_cpus=0,
+                scheduling_strategy=PlacementGroupSchedulingStrategy(
+                    self._head_pg,
+                    placement_group_bundle_index=slot,
+                    placement_group_capture_child_tasks=False,
+                ),
+            ).remote()
+            slot_to_actor[slot] = actor
+            slot_to_info_ref[slot] = actor.get_info.remote()
+
+        for slot in range(current, desired_num_actors):
+            _start_for_slot(slot)
 
-        pending = list(info_ref_to_actor.keys())
         deadline = time.time() + SCALE_ADD_TIMEOUT_S
         started = 0
 
-        while pending and time.time() < deadline:
-            remaining = len(pending)
+        while slot_to_info_ref and time.time() < deadline:
+            remaining = len(slot_to_info_ref)
             head_bundles = [{"CPU": 0, f"TPU-{self._tpu_type}-head": 1} for _ in range(remaining)]
             try:
                 request_resources(bundles=head_bundles)
             except Exception:
                 pass
 
-            done, pending = ray.wait(pending, num_returns=len(pending), timeout=SCALE_POLL_S)
+            pending_refs = list(slot_to_info_ref.values())
+            done, _ = ray.wait(pending_refs, num_returns=1, timeout=SCALE_POLL_S)
             if not done:
                 continue
 
-            for info_ref in done:
-                actor = info_ref_to_actor.pop(info_ref, None)
-                if actor is None:
+            for ref in done:
+                slot = next((s for s, r in slot_to_info_ref.items() if r == ref), None)
+                if slot is None:
                     continue
+                actor = slot_to_actor.get(slot)
+
+                slot_to_info_ref.pop(slot, None)
+
                 try:
-                    info = ray.get(info_ref, timeout=0)
+                    info = ray.get(ref, timeout=0)
+
                     self._actor_pool.append(ActorPoolMember(actor, info))
                     started += 1
-                    logger.info(f"Added actor {self.get_actor_name_from_actor_info(info)}")
+                    logger.info(f"Added actor {self.get_actor_name_from_actor_info(info)} (slot {slot})")
                 except Exception as e:
-                    logger.warning(f"SliceActor failed to start: {e}; killing actor")
+                    logger.warning(f"SliceActor for slot {slot} failed to start: {e}; killing and re-queuing")
                     try:
-                        ray.kill(actor)
+                        ray.kill(actor, no_restart=True)
                     except Exception:
                         pass
+                    _start_for_slot(slot)
 
-            logger.info(f"Started {started}/{num_to_add} slice actors so far")
+            logger.info(f"Started {started}/{desired_num_actors - current} slice actors so far")
 
-        if pending:
-            for info_ref in pending:
-                actor = info_ref_to_actor.get(info_ref)
-                if actor is not None:
-                    try:
-                        ray.kill(actor)
-                    except Exception:
-                        pass
-            logger.info(f"Started {started}/{num_to_add} slice actors (timed out for the rest)")
+        if slot_to_info_ref:
+            for _, actor in slot_to_actor.items():
+                try:
+                    if not any(m.actor == actor for m in self._actor_pool):
+                        ray.kill(actor, no_restart=True)
+                except Exception:
+                    pass
+            logger.info(f"Started {started}/{desired_num_actors - current} slice actors (timed out for the rest)")
+
+    def _ensure_head_pg(self, desired_num_actors: int) -> None:
+        """Ensure head placement group exists with the correct number of bundles.
+
+        Creates or recreates the head placement group to match the desired number of actors.
+        The placement group uses STRICT_SPREAD strategy to ensure SliceActors are distributed
+        across different nodes for optimal resource utilization and fault tolerance.
+
+        Each bundle in the placement group reserves one TPU head resource for a SliceActor.
+        If the current placement group doesn't match the target size, it's destroyed and
+        recreated with the correct number of bundles.
+
+        Args:
+            desired_num_actors: Target number of SliceActors (and placement group bundles)
+
+        Side Effects:
+            - Requests resources from Ray autoscaler for the new placement group
+            - Destroys existing placement group if size mismatch
+            - Updates _head_pg and _head_pg_target instance variables
+            - Blocks until placement group is ready
+
+        Note:
+            The placement group uses TPU-{tpu_type}-head resource labels for each bundle
+            to ensure proper resource allocation by the Ray autoscaler.
+        """
+        label = f"TPU-{self._tpu_type}-head"
+        if self._head_pg and self._head_pg_target == desired_num_actors:
+            return
+        if self._head_pg:
+            try:
+                remove_placement_group(self._head_pg)
+            except Exception:
+                pass
+            self._head_pg = None
+            self._head_pg_target = 0
+
+        bundles = [{"CPU": 0, label: 1} for _ in range(desired_num_actors)]
+        request_resources(bundles=bundles)
+        self._head_pg = placement_group(bundles, strategy="STRICT_SPREAD")
+        ray.get(self._head_pg.ready())
+        self._head_pg_target = desired_num_actors
+
+    def _destroy_head_pg(self) -> None:
+        """Destroy the current head placement group and reset tracking variables.
+
+        Removes the head placement group from Ray's cluster state and resets the
+        instance variables that track placement group state. This is typically
+        called during pool shutdown or when recreating placement groups with
+        different sizes.
+
+        The method uses best-effort cleanup - if the placement group removal fails
+        (e.g., due to Ray cluster issues), the error is logged but not propagated
+        to avoid disrupting the overall shutdown process.
+
+        Side Effects:
+            - Removes placement group from Ray cluster
+            - Sets _head_pg to None
+            - Resets _head_pg_target to 0
+            - Frees up cluster resources reserved by the placement group
+
+        Note:
+            This is a cleanup operation that should be called when the placement
+            group is no longer needed, such as during pool draining or before
+            recreating with a different size.
+        """
+        if self._head_pg:
+            try:
+                remove_placement_group(self._head_pg)
+            except Exception:
+                pass
+            self._head_pg = None
+            self._head_pg_target = 0
+
+    def drain_actor_pool(self) -> None:
+        """Shut down and remove all actors from the pool.
+
+        Attempts graceful shutdown first, then forcefully kills actors.
+        Clears the actor pool after draining.
+        """
+        super().drain_actor_pool()
+        self._destroy_head_pg()
