@@ -12,17 +12,20 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-from .dtypes import DTYPE_MAPPING, HAS_FLOAT8
+from .dtypes import DTYPE_MAPPING, DTYPE_TO_STRING_MAP, STRING_TO_DTYPE_MAP, put_dtype
 from .handler import PrecisionHandler
 from .loss_scaling import DynamicLossScale, LossScaleConfig, NoOpLossScale
 from .policy import Policy
 
 __all__ = (
     "DTYPE_MAPPING",
-    "HAS_FLOAT8",
+    "DTYPE_MAPPING",
+    "DTYPE_TO_STRING_MAP",
+    "STRING_TO_DTYPE_MAP",
     "DynamicLossScale",
     "LossScaleConfig",
     "NoOpLossScale",
     "Policy",
     "PrecisionHandler",
+    "put_dtype",
 )
